"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const NodeSchedule = require("node-schedule");
const Job_1 = require("./model/Job");
exports.Job = Job_1.Job;
const Errors_1 = require("./utils/Errors");
class Scheduler {
    constructor(reference, jobCompletionHandler) {
        if (!reference) {
            throw new Error(Errors_1.Errors.INVALID_ROOT_REFERENCE);
        }
        this.bonfireRef = reference;
        this.jobCompletionHandler = jobCompletionHandler;
        this.jobList = new Map();
        this.queueExistingJobs();
    }
    getRef() {
        return this.bonfireRef;
    }
    getPendingJobCount() {
        return this.jobList.size;
    }
    queueExistingJobs() {
        return __awaiter(this, void 0, void 0, function* () {
            const nodeSnapshot = yield this.bonfireRef.once('value');
            if (!nodeSnapshot.exists()) {
                return;
            }
            nodeSnapshot.forEach((jobSnapshot) => {
                if (!jobSnapshot.exists()) {
                    return false;
                }
                let job = Job_1.Job.fromJson(jobSnapshot.val());
                if (job.getScheduledDateTime().getTime() < Date.now()) {
                    this.jobCompletionHandler(job.getKey(), job);
                    jobSnapshot.ref.remove();
                    return false;
                }
                this.scheduleJob(job);
                return false;
            });
        });
    }
    getJobCallback(key) {
        return () => __awaiter(this, void 0, void 0, function* () {
            if (!this.jobCompletionHandler) {
                return;
            }
            const jobSnapshot = yield this.bonfireRef.child(key).once('value');
            if (!jobSnapshot.exists()) {
                return;
            }
            this.jobCompletionHandler(key, Job_1.Job.fromJson(jobSnapshot.val()));
            jobSnapshot.ref.remove();
        });
    }
    scheduleJob(job) {
        if (this.jobList.has(job.getKey())) {
            return;
        }
        const scheduledJob = NodeSchedule.scheduleJob(job.getKey(), job.getScheduledDateTime(), this.getJobCallback(job.getKey()));
        this.jobList.set(job.getKey(), scheduledJob);
    }
    schedule(job) {
        return __awaiter(this, void 0, void 0, function* () {
            if (job.getScheduledDateTime().getTime() < Date.now()) {
                throw new Error(Errors_1.Errors.SCHEDULED_IN_PAST);
            }
            let shouldCreateLocallyOnly = false;
            let jobSnapshot;
            jobSnapshot = yield this.bonfireRef.child(job.getKey()).once('value');
            if (jobSnapshot.exists()) {
                if (this.jobList.has(job.getKey())) {
                    return Job_1.Job.fromJson(jobSnapshot.val());
                }
                shouldCreateLocallyOnly = true;
            }
            this.scheduleJob(job);
            if (shouldCreateLocallyOnly) {
                return job;
            }
            yield jobSnapshot.ref.set(job.asJson());
            return job;
        });
    }
    cancel(key) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.jobList.has(key)) {
                return;
            }
            yield this.bonfireRef.child(key).remove();
            let job = this.jobList.get(key);
            job.cancel();
            this.jobList.delete(key);
        });
    }
}
exports.Scheduler = Scheduler;
